# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/monitor.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1esrc/ray/protobuf/monitor.proto\x12\x07ray.rpc\"\x16\n\x14GetRayVersionRequest\".\n\x12GetRayVersionReply\x12\x18\n\x07version\x18\x01 \x01(\tR\x07version\"4\n\x17\x44rainAndKillNodeRequest\x12\x19\n\x08node_ids\x18\x01 \x03(\x0cR\x07nodeIds\"<\n\x15\x44rainAndKillNodeReply\x12#\n\rdrained_nodes\x18\x02 \x03(\x0cR\x0c\x64rainedNodes2\xb6\x01\n\x11MonitorGcsService\x12K\n\rGetRayVersion\x12\x1d.ray.rpc.GetRayVersionRequest\x1a\x1b.ray.rpc.GetRayVersionReply\x12T\n\x10\x44rainAndKillNode\x12 .ray.rpc.DrainAndKillNodeRequest\x1a\x1e.ray.rpc.DrainAndKillNodeReplyB\x03\xf8\x01\x01\x62\x06proto3')



_GETRAYVERSIONREQUEST = DESCRIPTOR.message_types_by_name['GetRayVersionRequest']
_GETRAYVERSIONREPLY = DESCRIPTOR.message_types_by_name['GetRayVersionReply']
_DRAINANDKILLNODEREQUEST = DESCRIPTOR.message_types_by_name['DrainAndKillNodeRequest']
_DRAINANDKILLNODEREPLY = DESCRIPTOR.message_types_by_name['DrainAndKillNodeReply']
GetRayVersionRequest = _reflection.GeneratedProtocolMessageType('GetRayVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRAYVERSIONREQUEST,
  '__module__' : 'src.ray.protobuf.monitor_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetRayVersionRequest)
  })
_sym_db.RegisterMessage(GetRayVersionRequest)

GetRayVersionReply = _reflection.GeneratedProtocolMessageType('GetRayVersionReply', (_message.Message,), {
  'DESCRIPTOR' : _GETRAYVERSIONREPLY,
  '__module__' : 'src.ray.protobuf.monitor_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GetRayVersionReply)
  })
_sym_db.RegisterMessage(GetRayVersionReply)

DrainAndKillNodeRequest = _reflection.GeneratedProtocolMessageType('DrainAndKillNodeRequest', (_message.Message,), {
  'DESCRIPTOR' : _DRAINANDKILLNODEREQUEST,
  '__module__' : 'src.ray.protobuf.monitor_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DrainAndKillNodeRequest)
  })
_sym_db.RegisterMessage(DrainAndKillNodeRequest)

DrainAndKillNodeReply = _reflection.GeneratedProtocolMessageType('DrainAndKillNodeReply', (_message.Message,), {
  'DESCRIPTOR' : _DRAINANDKILLNODEREPLY,
  '__module__' : 'src.ray.protobuf.monitor_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DrainAndKillNodeReply)
  })
_sym_db.RegisterMessage(DrainAndKillNodeReply)

_MONITORGCSSERVICE = DESCRIPTOR.services_by_name['MonitorGcsService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _GETRAYVERSIONREQUEST._serialized_start=43
  _GETRAYVERSIONREQUEST._serialized_end=65
  _GETRAYVERSIONREPLY._serialized_start=67
  _GETRAYVERSIONREPLY._serialized_end=113
  _DRAINANDKILLNODEREQUEST._serialized_start=115
  _DRAINANDKILLNODEREQUEST._serialized_end=167
  _DRAINANDKILLNODEREPLY._serialized_start=169
  _DRAINANDKILLNODEREPLY._serialized_end=229
  _MONITORGCSSERVICE._serialized_start=232
  _MONITORGCSSERVICE._serialized_end=414
# @@protoc_insertion_point(module_scope)
