raise DeprecationWarning(
    "Ray SGD has been deprecated as of Ray 1.13. For distributed "
    "deep learning on Ray please use Ray AI Runtime (Ray AIR) "
    "instead ("
    "https://docs.ray.io/en/master/ray-air/getting-started.html)."
)
