# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/runtime_env_common.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)src/ray/protobuf/runtime_env_common.proto\x12\x07ray.rpc\"`\n\x0eRuntimeEnvUris\x12&\n\x0fworking_dir_uri\x18\x01 \x01(\tR\rworkingDirUri\x12&\n\x0fpy_modules_uris\x18\x02 \x03(\tR\rpyModulesUris\"k\n\x10RuntimeEnvConfig\x12\x32\n\x15setup_timeout_seconds\x18\x01 \x01(\x05R\x13setupTimeoutSeconds\x12#\n\reager_install\x18\x02 \x01(\x08R\x0c\x65\x61gerInstall\"\xbc\x01\n\x0eRuntimeEnvInfo\x12\x34\n\x16serialized_runtime_env\x18\x01 \x01(\tR\x14serializedRuntimeEnv\x12+\n\x04uris\x18\x02 \x01(\x0b\x32\x17.ray.rpc.RuntimeEnvUrisR\x04uris\x12G\n\x12runtime_env_config\x18\x03 \x01(\x0b\x32\x19.ray.rpc.RuntimeEnvConfigR\x10runtimeEnvConfig\"\xdf\x01\n\x0fRuntimeEnvState\x12\x1f\n\x0bruntime_env\x18\x01 \x01(\tR\nruntimeEnv\x12\x17\n\x07ref_cnt\x18\x02 \x01(\x03R\x06refCnt\x12\x1d\n\x07success\x18\x03 \x01(\x08H\x00R\x07success\x88\x01\x01\x12\x19\n\x05\x65rror\x18\x04 \x01(\tH\x01R\x05\x65rror\x88\x01\x01\x12-\n\x10\x63reation_time_ms\x18\x05 \x01(\x03H\x02R\x0e\x63reationTimeMs\x88\x01\x01\x42\n\n\x08_successB\x08\n\x06_errorB\x13\n\x11_creation_time_msB\x1d\n\x18io.ray.runtime.generated\xf8\x01\x01\x62\x06proto3')



_RUNTIMEENVURIS = DESCRIPTOR.message_types_by_name['RuntimeEnvUris']
_RUNTIMEENVCONFIG = DESCRIPTOR.message_types_by_name['RuntimeEnvConfig']
_RUNTIMEENVINFO = DESCRIPTOR.message_types_by_name['RuntimeEnvInfo']
_RUNTIMEENVSTATE = DESCRIPTOR.message_types_by_name['RuntimeEnvState']
RuntimeEnvUris = _reflection.GeneratedProtocolMessageType('RuntimeEnvUris', (_message.Message,), {
  'DESCRIPTOR' : _RUNTIMEENVURIS,
  '__module__' : 'src.ray.protobuf.runtime_env_common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RuntimeEnvUris)
  })
_sym_db.RegisterMessage(RuntimeEnvUris)

RuntimeEnvConfig = _reflection.GeneratedProtocolMessageType('RuntimeEnvConfig', (_message.Message,), {
  'DESCRIPTOR' : _RUNTIMEENVCONFIG,
  '__module__' : 'src.ray.protobuf.runtime_env_common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RuntimeEnvConfig)
  })
_sym_db.RegisterMessage(RuntimeEnvConfig)

RuntimeEnvInfo = _reflection.GeneratedProtocolMessageType('RuntimeEnvInfo', (_message.Message,), {
  'DESCRIPTOR' : _RUNTIMEENVINFO,
  '__module__' : 'src.ray.protobuf.runtime_env_common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RuntimeEnvInfo)
  })
_sym_db.RegisterMessage(RuntimeEnvInfo)

RuntimeEnvState = _reflection.GeneratedProtocolMessageType('RuntimeEnvState', (_message.Message,), {
  'DESCRIPTOR' : _RUNTIMEENVSTATE,
  '__module__' : 'src.ray.protobuf.runtime_env_common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RuntimeEnvState)
  })
_sym_db.RegisterMessage(RuntimeEnvState)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030io.ray.runtime.generated\370\001\001'
  _RUNTIMEENVURIS._serialized_start=54
  _RUNTIMEENVURIS._serialized_end=150
  _RUNTIMEENVCONFIG._serialized_start=152
  _RUNTIMEENVCONFIG._serialized_end=259
  _RUNTIMEENVINFO._serialized_start=262
  _RUNTIMEENVINFO._serialized_end=450
  _RUNTIMEENVSTATE._serialized_start=453
  _RUNTIMEENVSTATE._serialized_end=676
# @@protoc_insertion_point(module_scope)
