# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import runtime_env_agent_pb2 as src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2


class RuntimeEnvServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetOrCreateRuntimeEnv = channel.unary_unary(
                '/ray.rpc.RuntimeEnvService/GetOrCreateRuntimeEnv',
                request_serializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetOrCreateRuntimeEnvRequest.SerializeToString,
                response_deserializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetOrCreateRuntimeEnvReply.FromString,
                )
        self.DeleteRuntimeEnvIfPossible = channel.unary_unary(
                '/ray.rpc.RuntimeEnvService/DeleteRuntimeEnvIfPossible',
                request_serializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.DeleteRuntimeEnvIfPossibleRequest.SerializeToString,
                response_deserializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.DeleteRuntimeEnvIfPossibleReply.FromString,
                )
        self.GetRuntimeEnvsInfo = channel.unary_unary(
                '/ray.rpc.RuntimeEnvService/GetRuntimeEnvsInfo',
                request_serializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetRuntimeEnvsInfoRequest.SerializeToString,
                response_deserializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetRuntimeEnvsInfoReply.FromString,
                )


class RuntimeEnvServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetOrCreateRuntimeEnv(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteRuntimeEnvIfPossible(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRuntimeEnvsInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RuntimeEnvServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetOrCreateRuntimeEnv': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOrCreateRuntimeEnv,
                    request_deserializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetOrCreateRuntimeEnvRequest.FromString,
                    response_serializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetOrCreateRuntimeEnvReply.SerializeToString,
            ),
            'DeleteRuntimeEnvIfPossible': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteRuntimeEnvIfPossible,
                    request_deserializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.DeleteRuntimeEnvIfPossibleRequest.FromString,
                    response_serializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.DeleteRuntimeEnvIfPossibleReply.SerializeToString,
            ),
            'GetRuntimeEnvsInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRuntimeEnvsInfo,
                    request_deserializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetRuntimeEnvsInfoRequest.FromString,
                    response_serializer=src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetRuntimeEnvsInfoReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ray.rpc.RuntimeEnvService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RuntimeEnvService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetOrCreateRuntimeEnv(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ray.rpc.RuntimeEnvService/GetOrCreateRuntimeEnv',
            src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetOrCreateRuntimeEnvRequest.SerializeToString,
            src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetOrCreateRuntimeEnvReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteRuntimeEnvIfPossible(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ray.rpc.RuntimeEnvService/DeleteRuntimeEnvIfPossible',
            src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.DeleteRuntimeEnvIfPossibleRequest.SerializeToString,
            src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.DeleteRuntimeEnvIfPossibleReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRuntimeEnvsInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ray.rpc.RuntimeEnvService/GetRuntimeEnvsInfo',
            src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetRuntimeEnvsInfoRequest.SerializeToString,
            src_dot_ray_dot_protobuf_dot_runtime__env__agent__pb2.GetRuntimeEnvsInfoReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
