# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/logging.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1esrc/ray/protobuf/logging.proto\x12\x07ray.rpc\"\xb0\x01\n\x08LogBatch\x12\x0e\n\x02ip\x18\x01 \x01(\tR\x02ip\x12\x10\n\x03pid\x18\x02 \x01(\tR\x03pid\x12\x15\n\x06job_id\x18\x03 \x01(\tR\x05jobId\x12\x19\n\x08is_error\x18\x04 \x01(\x08R\x07isError\x12\x14\n\x05lines\x18\x05 \x03(\tR\x05lines\x12\x1d\n\nactor_name\x18\x06 \x01(\tR\tactorName\x12\x1b\n\ttask_name\x18\x07 \x01(\tR\x08taskNameB\x03\xf8\x01\x01\x62\x06proto3')



_LOGBATCH = DESCRIPTOR.message_types_by_name['LogBatch']
LogBatch = _reflection.GeneratedProtocolMessageType('LogBatch', (_message.Message,), {
  'DESCRIPTOR' : _LOGBATCH,
  '__module__' : 'src.ray.protobuf.logging_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.LogBatch)
  })
_sym_db.RegisterMessage(LogBatch)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _LOGBATCH._serialized_start=44
  _LOGBATCH._serialized_end=220
# @@protoc_insertion_point(module_scope)
