raise DeprecationWarning(
    "`ray.train.callbacks.results_preprocessors` and the `ray.train.Trainer` API are "
    "deprecated in Ray "
    "2.0, and are replaced by Ray AI Runtime (Ray AIR). Ray AIR "
    "(https://docs.ray.io/en/latest/ray-air/getting-started.html) "
    "provides greater functionality and a unified API "
    "compared to the old Ray Train API. "
)
